# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Common stuff"

import abc
import functools
import inspect
import logging
import os
import os.path as p
import pprint
import re
import shutil
import signal
import subprocess as subp
import sys
import threading
from collections import Counter
from tempfile import NamedTemporaryFile
from threading import Timer
from typing import Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

import six

_logger = logging.getLogger(__name__)

ON_WINDOWS = os.name == "nt"
ON_LINUX = sys.platform == "linux"
ON_MAC = sys.platform == "darwin"


def setupLogging(stream, level):  # pragma: no cover
    "Setup logging according to the command line parameters"
    if isinstance(stream, six.string_types):
        _stream = open(stream, "a")
    else:
        _stream = stream

    handler = logging.StreamHandler(_stream)
    handler.formatter = logging.Formatter(
        "%(levelname)-7s | %(asctime)s | "
        + "%(name)s @ %(funcName)s():%(lineno)d %(threadName)s "
        + "|\t%(message)s",
        datefmt="%H:%M:%S",
    )

    logging.root.addHandler(handler)
    logging.root.setLevel(level)

    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("pynvim").setLevel(logging.WARNING)
    logging.getLogger("matplotlib").setLevel(logging.INFO)


# From here: http://stackoverflow.com/a/8536476/1672783
def terminateProcess(pid):
    "Terminate a process given its PID"

    if ON_WINDOWS:
        import ctypes  # pylint: disable=import-outside-toplevel

        process_terminate = 1
        handle = ctypes.windll.kernel32.OpenProcess(process_terminate, False, pid)
        ctypes.windll.kernel32.TerminateProcess(handle, -1)
        ctypes.windll.kernel32.CloseHandle(handle)
    else:
        os.kill(pid, signal.SIGTERM)


def isProcessRunning(pid):
    "Checks if a process is running given its PID"

    if ON_WINDOWS:
        return _isProcessRunningOnWindows(pid)

    return _isProcessRunningOnPosix(pid)


def _isProcessRunningOnPosix(pid):
    "Checks if a given PID is runnning under POSIX OSs"
    try:
        os.kill(pid, 0)
    except OSError:
        return False

    return True


def _isProcessRunningOnWindows(pid):
    """
    Enumerates active processes as seen under windows Task Manager on Win
    NT/2k/XP using PSAPI.dll (new api for processes) and using ctypes.Use it as
    you please.

    Based on information from
    http://support.microsoft.com/default.aspx?scid=KB;EN-US;Q175030&ID=KB;EN-US;Q175030

    By Eric Koome email ekoome@yahoo.com
    license GPL

    (adapted from code found at
    http://code.activestate.com/recipes/305279-getting-process-information-on-windows/)
    """
    from ctypes import (  # pylint: disable=import-outside-toplevel
        windll,
        c_ulong,
        sizeof,
        byref,
    )

    # PSAPI.DLL
    psapi = windll.psapi

    arr = c_ulong * 256
    list_of_pids = arr()
    cb = sizeof(list_of_pids)  # pylint: disable=invalid-name
    cb_needed = c_ulong()

    # Call Enumprocesses to get hold of process id's
    psapi.EnumProcesses(byref(list_of_pids), cb, byref(cb_needed))

    # Number of processes returned
    number_of_pids = int(cb_needed.value / sizeof(c_ulong()))

    pid_list = list(list_of_pids)[:number_of_pids]

    return int(pid) in pid_list


if not hasattr(p, "samefile"):

    def _samefile(file1, file2):
        """
        Emulated version of os.path.samefile. This is needed for Python
        2.7 running on Windows (at least on Appveyor CI)
        """

        return os.stat(file1) == os.stat(file2)


else:
    _samefile = p.samefile  # pylint: disable=invalid-name

samefile = _samefile  # pylint: disable=invalid-name


def removeDuplicates(seq):
    """
    Fast removal of duplicates within an iterable
    """
    seen = set()
    seen_add = seen.add

    return [x for x in seq if not (x in seen or seen_add(x))]


# Copied from ycmd
def toBytes(value):  # pragma: no cover
    """
    Consistently returns the new bytes() type from python-future.
    Assumes incoming strings are either UTF-8 or unicode (which is
    converted to UTF-8).
    """

    if not value:
        return bytes()

    # This is tricky. On py2, the bytes type from builtins (from python-future) is
    # a subclass of str. So all of the following are true:
    #   isinstance(str(), bytes)
    #   isinstance(bytes(), str)
    # But they don't behave the same in one important aspect: iterating over a
    # bytes instance yields ints, while iterating over a (raw, py2) str yields
    # chars. We want consistent behavior so we force the use of bytes().

    if isinstance(value, bytes):
        return value

    # This is meant to catch Python 2's native str type.

    if isinstance(value, bytes):
        return bytes(value, encoding="utf8")

    if isinstance(value, str):
        # On py2, with `from builtins import *` imported, the following is true:
        #
        #   bytes(str(u'abc'), 'utf8') == b"b'abc'"
        #
        # Obviously this is a bug in python-future. So we work around it. Also filed
        # upstream at: https://github.com/PythonCharmers/python-future/issues/193
        # We can't just return value.encode('utf8') on both py2 & py3 because on
        # py2 that *sometimes* returns the built-in str type instead of the newbytes
        # type from python-future.

        if six.PY2:
            return bytes(value.encode("utf8"), encoding="utf8")

        return bytes(value, encoding="utf8")

    # This is meant to catch `int` and similar non-string/bytes types.

    return toBytes(str(value))


def getTemporaryFilename(name):
    """
    Gets a temporary filename following the format 'hdl_checker_pid<>.log' on Linux
    and 'hdl_checker_pid<>_<unique>.log' on Windows
    """
    try:
        name, suffix = name.split(".")
    except ValueError:
        suffix = None

    basename = "hdl_checker_" + name + "_pid{}".format(os.getpid())

    if ON_WINDOWS:
        return NamedTemporaryFile(
            prefix=basename + "_", suffix="." + (suffix or "log"), delete=False
        ).name

    return p.join(p.sep, "tmp", basename + "." + (suffix or "log"))


def isFileReadable(path):
    # type: (str) -> bool
    """
    Checks if a given file is readable
    """
    try:
        open(path, "r").close()

        return True
    except IOError:
        return False


def runShellCommand(cmd_with_args, shell=False, env=None, cwd=None):
    # type: (Union[Tuple[str], List[str]], bool, Optional[Dict], Optional[str]) -> Iterable[str]
    """
    Runs a shell command and handles stdout catching
    """
    _logger.debug(" ".join(cmd_with_args))

    try:
        return (
            subp.check_output(
                cmd_with_args,
                stderr=subp.STDOUT,
                shell=shell,
                env=env or os.environ,
                cwd=cwd,
            )
            .decode(errors="replace")
            .splitlines()
        )
    except subp.CalledProcessError as exc:
        stdout = tuple(exc.output.decode(errors="replace").splitlines())
        _logger.debug(
            "Command '%s' failed with error code %d.\nStdout:\n%s",
            cmd_with_args,
            exc.returncode,
            "\n".join(stdout),
        )
        return stdout
    except OSError as exc:
        _logger.debug("Command '%s' failed with %s", cmd_with_args, exc)
        raise


def removeIfExists(filename):
    # type: (str) -> bool
    "Removes filename using os.remove and catches the exception if that fails"
    try:
        os.remove(filename)
        _logger.debug("Removed %s", filename)
        return True
    except OSError:
        _logger.debug("Failed to remove %s", filename)
        return False


def removeDirIfExists(dirname):
    # type: (str) -> bool
    """
    Removes the directory dirname using shutil.rmtree and catches the exception
    if that fails
    """
    try:
        shutil.rmtree(dirname)
        _logger.debug("Removed %s", dirname)
        return True
    except OSError:
        _logger.debug("Failed to remove %s", dirname)
        return False


class HashableByKey(object):  # pylint: disable=useless-object-inheritance
    """
    Implements hash and comparison operators properly across Python 2 and 3
    """

    __metaclass__ = abc.ABCMeta

    @property
    @abc.abstractmethod
    def __hash_key__(self):
        """ Implement this attribute to use it for hashing and comparing"""

    def __hash__(self):
        try:
            return hash(self.__hash_key__)
        except:  # pragma: no cover
            print("Couldn't hash %s" % repr(self.__hash_key__))
            raise

    def __eq__(self, other):
        """Overrides the default implementation"""

        if isinstance(other, self.__class__):
            return self.__hash_key__ == other.__hash_key__

        return NotImplemented  # pragma: no cover

    def __ne__(self, other):  # pragma: no cover
        """Overrides the default implementation (unnecessary in Python 3)"""
        result = self.__eq__(other)

        if result is not NotImplemented:
            return not result

        return NotImplemented


def logCalls(func):  # pragma: no cover
    # type: (Callable) -> Callable
    "Decorator to Log calls to func"

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # type: (...) -> Callable
        _str = "%s(%s, %s)" % (func.__name__, args, pprint.pformat(kwargs))
        try:
            result = func(self, *args, **kwargs)
            _logger.debug("%s => %s", _str, repr(result))

            return result
        except:
            _logger.exception("Failed to run %s", _str)
            raise

    return wrapper


T = TypeVar("T")  # pylint: disable=invalid-name


def getMostCommonItem(items):
    # type: (Iterable[T]) -> T
    """
    Gets the most common item on an interable of items
    """
    data = Counter(items)
    return max(items, key=data.get)


def readFile(path):
    "Wrapper around open().read() that return \n for new lines"
    return open(path, mode="r", newline="\n", errors="replace").read()


REPO_URL = "https://github.com/suoto/hdl_checker"
_TAGS = re.compile(r"^\w+\s+refs\/tags\/v(?P<tag>(?:\d+\.){2}\d+)", flags=re.MULTILINE)

VersionFormat = Tuple[int, ...]


def _getLatestReleaseVersion():
    # type: () -> Optional[VersionFormat]
    """
    Return the latest tag from https://github.com/suoto/hdl_checker, striping
    the leading 'v' (so that v1.0.0 becomes simply 1.0.0). If the connection to
    the URL fails, return None
    """
    proc = subp.Popen(
        ["git", "ls-remote", "--tags", REPO_URL],
        env={"GIT_TERMINAL_PROMPT": "0"},
        stdout=subp.PIPE,
        stderr=subp.PIPE,
    )

    timer = Timer(5, proc.kill)
    timer.start()

    try:
        stdout, stderr = proc.communicate()
    finally:
        timer.cancel()

    if not stdout or stderr:
        _logger.info(
            "Couldn't fetch latest tag from %s: '%s'", REPO_URL, stderr.decode()
        )
        return None

    tags = tuple(
        tuple(int(x) for x in tag.split(".")) for tag in _TAGS.findall(stdout.decode())
    )

    if tags:
        return sorted(tags)[-1]

    _logger.warning("Unable to get version from '%s'", stdout)
    return None


_VERSION_FORMAT = re.compile(r"^\d+\.\d+\.\d+$")


def onNewReleaseFound(func):
    # type: (Callable[[str], None]) -> None
    """
    Checks if a new release is out and calls func if the running an older
    version
    """
    from hdl_checker import (  # pylint: disable=import-outside-toplevel
        __version__ as current,
    )

    # When installing via pip from github, versioneer will report the current
    # version as 0+unknown, in which case we won't notify
    if not _VERSION_FORMAT.match(current):
        return

    latest = _getLatestReleaseVersion()

    if not latest:
        return

    _logger.debug("Current version is %s, latest is %s", current, latest)

    if latest > tuple(int(x) for x in current.split(".")):
        func(
            "HDL Checker version {} is out! (current version is {})".format(
                ".".join(map(str, latest)), current
            )
        )


ENABLE_DEBOUNCE = True

# Copied from pyls (see
# https://github.com/palantir/python-language-server/blob/d81c7ba14d54b8e52192b0e00cbb4dacbb6f414d/pyls/_utils.py#L22-L47)
def debounce(interval_s, keyed_by=None):
    """Debounce calls to this function until interval_s seconds have passed."""

    def wrapper(func):
        timers = {}
        lock = threading.Lock()

        @functools.wraps(func)
        def debounced(*args, **kwargs):
            key = (
                inspect.signature(func).bind(*args, **kwargs).arguments[keyed_by]
                if keyed_by
                else None
            )

            # Could not find a way to disable debouncing whilist testing, so
            # we'll use a module var
            if not ENABLE_DEBOUNCE:
                return func(*args, **kwargs)

            def run():
                with lock:
                    del timers[key]
                return func(*args, **kwargs)

            with lock:
                old_timer = timers.get(key)
                if old_timer:
                    old_timer.cancel()

                timer = threading.Timer(interval_s, run)
                timers[key] = timer
                timer.start()

        return debounced

    return wrapper



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring

import asyncio
import json
import logging
import os
import os.path as p
from threading import Thread
from typing import Any, List, Optional, Union

import unittest2  # type: ignore
from mock import patch
from pygls import features, uris
from pygls.server import LanguageServer
from pygls.types import (
    ClientCapabilities,
    DiagnosticSeverity,
    DidChangeTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    HoverAbstract,
    HoverParams,
    InitializeParams,
    MarkupKind,
    Position,
    PublishDiagnosticsAbstract,
    Range,
    ReferenceContext,
    ReferenceParams,
    TextDocumentClientCapabilities,
    TextDocumentContentChangeEvent,
    TextDocumentIdentifier,
    TextDocumentItem,
    TextDocumentPositionParams,
    VersionedTextDocumentIdentifier,
)
from tabulate import tabulate

from hdl_checker.tests import getTestTempPath, setupTestSuport, toCheckerDiagnostic

import hdl_checker
from hdl_checker import DEFAULT_LIBRARY, lsp
from hdl_checker.diagnostics import CheckerDiagnostic
from hdl_checker.parsers.elements.dependency_spec import RequiredDesignUnit
from hdl_checker.parsers.elements.design_unit import (
    DesignUnitType,
    VerilogDesignUnit,
    VhdlDesignUnit,
)
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import Location
from hdl_checker.utils import ON_WINDOWS

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.join(TEST_TEMP_PATH, "test_project")
LSP_REQUEST_TIMEOUT = 3

_CLIENT_CAPABILITIES = ClientCapabilities(
    text_document=TextDocumentClientCapabilities(
        synchronization=None,  # type: ignore
        completion=None,  # type: ignore
        hover=HoverAbstract(
            dynamic_registration=False, content_format=[MarkupKind.PlainText,],
        ),
        signature_help=None,  # type:ignore
        references=None,  # type:ignore
        document_highlight=None,  # type:ignore
        document_symbol=None,  # type:ignore
        formatting=None,  # type:ignore
        range_formatting=None,  # type:ignore
        on_type_formatting=None,  # type:ignore
        definition=None,  # type:ignore
        type_definition=None,  # type:ignore
        implementation=None,  # type:ignore
        code_action=None,  # type:ignore
        code_lens=None,  # type:ignore
        document_link=None,  # type:ignore
        color_provider=None,  # type:ignore
        rename=None,  # type:ignore
        publish_diagnostics=PublishDiagnosticsAbstract(related_information=True),
        folding_range=None,  # type:ignore
    )
)

if ON_WINDOWS:
    TEST_PROJECT = TEST_PROJECT.lower()


class _LspHelper(unittest2.TestCase):
    def _createClientServerPair(self, params: Optional[InitializeParams]):
        # pylint: disable=attribute-defined-outside-init
        setupTestSuport(TEST_TEMP_PATH)
        _logger.debug("Creating server")

        # Client to Server pipe
        csr, csw = os.pipe()
        # Server to client pipe
        scr, scw = os.pipe()

        # Setup server
        self.server = lsp.HdlCheckerLanguageServer()
        lsp.setupLanguageServerFeatures(self.server)

        self.server_diagnostics = []  # type: ignore

        @self.server.feature(features.TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)
        def serverPublishDiagnosticsHandler(diag):  # pylint: disable=unused-variable
            _logger.info("Publishing diagnostic: %s", diag)
            self.server_diagnostics.append(diag)

        self.server_thread = Thread(
            target=self.server.start_io,
            args=(os.fdopen(csr, "rb"), os.fdopen(scw, "wb")),
        )

        self.server_thread.daemon = True
        self.server_thread.name = "server"
        self.server_thread.start()

        # Add thread id to the server (just for testing)
        self.server.thread_id = self.server_thread.ident  # type: ignore

        # Setup client (client doesn't need the HDL checker stuff)
        self.client = LanguageServer(asyncio.new_event_loop())
        self.client_messages = []  # type: ignore
        self.client_diagnostics = []  # type: ignore

        @self.client.feature(features.WINDOW_SHOW_MESSAGE)
        def clientShowMessageHandler(*args):  # pylint: disable=unused-variable
            type_ = DiagnosticSeverity(args[0][0])
            msg = args[0][1]
            _logger.info("[Client] [%s]: %s", type_, msg)
            self.client_messages.append((type_, msg))

        @self.client.feature(features.TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)
        def clientPublishDiagnosticsHandler(diag):  # pylint: disable=unused-variable
            _logger.info("Publishing diagnostic: %s", diag)
            self.client_diagnostics.append(diag)

        self.client_thread = Thread(
            target=self.client.start_io,
            args=(os.fdopen(scr, "rb"), os.fdopen(csw, "wb")),
        )

        self.client_thread.daemon = True
        self.client_thread.name = "client"
        self.client_thread.start()

        if params is None:
            _logger.info("No parameters given, won't initialize server")
            return

        _logger.info("Sending initialize request")
        self.client.lsp.send_request(features.INITIALIZE, params).result(
            LSP_REQUEST_TIMEOUT
        )

        _logger.info("Sending initialized request")
        self.client.lsp.send_request(features.INITIALIZED).result(LSP_REQUEST_TIMEOUT)
        _logger.info("Client server setup complete")

    def tearDown(self):  # pylint:disable=inconsistent-return-statements
        if self.__class__ is _LspHelper:
            self.assertIsNone(getattr(self, "server", None))
            self.assertIsNone(getattr(self, "client", None))
            return unittest2.skip("Won't run this on %s" % self.__class__)
        _logger.info("#" * 100)
        _logger.info("Shutting down server")
        shutdown_response = self.client.lsp.send_request(features.SHUTDOWN).result(
            LSP_REQUEST_TIMEOUT
        )
        self.client.lsp.notify(features.EXIT)
        self.assertIsNone(shutdown_response)
        # pygls server has cleanup routines run when the interpreter is
        # exiting, at which point the client will be long gone. This results in
        # RuntimeError: cannot schedule new futures after shutdown. To work
        # around this, we're accessing its itnernal event to force it to exit
        self.client._stop_event.set()
        self.server_thread.join()
        self.client_thread.join()

    def checkLintFileOnMethod(
        self,
        params: Union[
            DidOpenTextDocumentParams,
            DidSaveTextDocumentParams,
            DidChangeTextDocumentParams,
        ],
        expected_diags=List[CheckerDiagnostic],
    ):
        """
        Generic method to check diagnostics reported are correct
        """
        filename = uris.to_fs_path(params.textDocument.uri)
        _logger.info("Checking lint of file %s, event is %s", filename, type(params))
        assert isinstance(
            params,
            (
                DidOpenTextDocumentParams,
                DidSaveTextDocumentParams,
                DidChangeTextDocumentParams,
            ),
        )

        if isinstance(params, DidOpenTextDocumentParams):
            method = features.TEXT_DOCUMENT_DID_OPEN
        elif isinstance(params, DidSaveTextDocumentParams):
            method = features.TEXT_DOCUMENT_DID_SAVE
        else:
            method = features.TEXT_DOCUMENT_DID_CHANGE

        self.assertFalse(
            self.client_diagnostics, "Client shoult not have diagnostics by now"
        )

        _logger.info("Patching %s with %s", "getMessagesByPath", list(expected_diags))
        with patch.object(
            self.server.checker, "getMessagesByPath", return_value=list(expected_diags),
        ):
            hdl_checker.utils.ENABLE_DEBOUNCE = False
            try:
                self.client.lsp.send_request(method, params).result(LSP_REQUEST_TIMEOUT)
            finally:
                hdl_checker.utils.ENABLE_DEBOUNCE = True

            self.assertTrue(
                self.client_diagnostics, "Expected client to have diagnostics"
            )
            diags: List[CheckerDiagnostic] = []
            while self.client_diagnostics:
                diag = self.client_diagnostics.pop()
                diags += list(toCheckerDiagnostic(diag[0], diag[1]))

            _logger.info("Expected: %d => %s", len(expected_diags), expected_diags)
            _logger.info("Got:      %d => %s", len(diags), diags)
            self.assertFalse(set(expected_diags) - set(diags))

    def _runDidOpenCheck(self, source: Optional[Path]):
        self.checkLintFileOnMethod(
            DidOpenTextDocumentParams(
                TextDocumentItem(
                    uris.from_fs_path(source), language_id="vhdl", version=0, text="",
                )
            ),
            [
                CheckerDiagnostic(
                    text="testing clk en gen",
                    filename=source,
                    line_number=0,
                    column_number=0,
                ),
            ],
        )

    def _runDidSaveCheck(self, source: Optional[Path]):
        self.checkLintFileOnMethod(
            DidSaveTextDocumentParams(
                text_document=TextDocumentIdentifier(uris.from_fs_path(source)),
                text="Hello",
            ),
            [
                CheckerDiagnostic(
                    text="testing clk en gen",
                    filename=source,
                    line_number=0,
                    column_number=0,
                ),
            ],
        )

    def _runDidChangeCheck(self, source: Optional[Path]):
        self.checkLintFileOnMethod(
            DidChangeTextDocumentParams(
                VersionedTextDocumentIdentifier(uris.from_fs_path(source), version=1,),
                [
                    TextDocumentContentChangeEvent(
                        range=Range(Position(0, 0), Position(1, 1))
                    ),
                ],
            ),
            [
                CheckerDiagnostic(
                    text="post change diag",
                    filename=source,
                    line_number=1,
                    column_number=2,
                ),
            ],
        )

    def test_LintFileOnOpen(self):  # pylint: disable=inconsistent-return-statements
        if self.__class__ is _LspHelper:
            return unittest2.skip("Won't run this on %s" % self.__class__)
        _logger.info("#" * 100)
        self._runDidOpenCheck(p.join(TEST_PROJECT, "another_library", "foo.vhd"))

    def test_LintFileWhenSaving(self):  # pylint: disable=inconsistent-return-statements
        if self.__class__ is _LspHelper:
            return unittest2.skip("Won't run this on %s" % self.__class__)
        _logger.info("#" * 100)
        self._runDidSaveCheck(
            p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")
        )

    def test_LintFileOnChange(self):  # pylint: disable=inconsistent-return-statements
        if self.__class__ is _LspHelper:
            return unittest2.skip("Won't run this on %s" % self.__class__)
        _logger.info("#" * 100)
        self._runDidOpenCheck(
            p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
        )
        self._runDidChangeCheck(
            p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
        )


class TestRootUriNoProjectFile(_LspHelper):
    def setUp(self):
        _logger.info("#" * 100)
        setupTestSuport(TEST_TEMP_PATH)
        _logger.debug("Creating server")

        self.patches = (
            patch.object(
                hdl_checker.config_generators.base_generator.BaseGenerator, "generate",
            ),
            patch("hdl_checker.core.json.dump", spec=json.dump),
        )

        for _patch in self.patches:
            _patch.start()

        self._createClientServerPair(
            InitializeParams(
                process_id=1234,
                capabilities=_CLIENT_CAPABILITIES,
                root_uri=uris.from_fs_path(TEST_PROJECT),
            )
        )

    def tearDown(self):
        _logger.info("#" * 100)
        _logger.info("Shutting down server")
        for _patch in self.patches:
            _patch.stop()
        super(TestRootUriNoProjectFile, self).tearDown()

    def test_SearchesForFilesOnInitialization(
        self,
    ):  # pylint: disable=no-self-use,invalid-name
        _logger.info("#" * 100)
        lsp.SimpleFinder.generate.assert_called_once()  # pylint: disable=no-member
        #  Will get called twice
        hdl_checker.core.json.dump.assert_called()  # pylint: disable=no-member


class TestOldStyleProjectFile(_LspHelper):
    def setUp(self):
        _logger.info("#" * 100)
        setupTestSuport(TEST_TEMP_PATH)
        _logger.debug("Creating server")

        self._createClientServerPair(
            InitializeParams(
                process_id=1235,
                capabilities=_CLIENT_CAPABILITIES,
                root_uri=uris.from_fs_path(TEST_PROJECT),
                initialization_options={"project_file": "vimhdl.prj"},
            ),
        )


class TestNonExistingProjectFile(_LspHelper):
    def setUp(self):
        _logger.info("#" * 100)
        setupTestSuport(TEST_TEMP_PATH)
        _logger.debug("Creating server")
        self.project_file = "__some_project_file.prj"
        self.assertFalse(p.exists(self.project_file))

        self._createClientServerPair(
            InitializeParams(
                process_id=1236,
                capabilities=_CLIENT_CAPABILITIES,
                root_uri=uris.from_fs_path(TEST_PROJECT),
                initialization_options={"project_file": self.project_file},
            ),
        )


class TestNoRootNoProjectFile(_LspHelper):
    def setUp(self):
        _logger.info("#" * 100)
        setupTestSuport(TEST_TEMP_PATH)
        _logger.debug("Creating server")
        self.project_file = "__some_project_file.prj"
        self.assertFalse(p.exists(self.project_file))

        self._createClientServerPair(
            InitializeParams(
                process_id=1236,
                capabilities=_CLIENT_CAPABILITIES,
                root_uri=None,
                initialization_options={"project_file": None},
            ),
        )


class TestNoRootWithProjectFile(_LspHelper):
    def setUp(self):
        setupTestSuport(TEST_TEMP_PATH)
        _logger.debug("Creating server")
        self.project_file = "__some_project_file.prj"
        self.assertFalse(p.exists(self.project_file))

        self._createClientServerPair(
            InitializeParams(
                process_id=1237,
                capabilities=_CLIENT_CAPABILITIES,
                root_uri=None,
                initialization_options={
                    "project_file": p.join(TEST_PROJECT, "vimhdl.prj")
                },
            ),
        )


class TestValidProject(_LspHelper):
    def setUp(self):
        _logger.info("#" * 100)
        setupTestSuport(TEST_TEMP_PATH)
        _logger.debug("Creating server")
        self.project_file = "__some_project_file.prj"
        self.assertFalse(p.exists(self.project_file))

        self._createClientServerPair(
            InitializeParams(
                process_id=1238,
                capabilities=_CLIENT_CAPABILITIES,
                root_uri=uris.from_fs_path(TEST_PROJECT),
                initialization_options={"project_file": "config.json"},
            ),
        )

    def runTestBuildSequenceTable(self, tablefmt):
        _logger.debug("############################################")
        very_common_pkg = Path(
            p.join(TEST_PROJECT, "basic_library", "very_common_pkg.vhd")
        )
        clk_en_generator = Path(
            p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
        )

        expected = [
            "Build sequence for %s is" % str(clk_en_generator),
            "",
            tabulate(
                [
                    (1, "basic_library", str(very_common_pkg)),
                    (2, DEFAULT_LIBRARY.name, str(clk_en_generator)),
                ],
                headers=("#", "Library", "Path"),
                tablefmt=tablefmt,
            ),
        ]

        try:
            got = self.server.getBuildSequenceForHover(clk_en_generator)
            self.assertEqual(got, "\n".join(expected))
        except:
            _logger.error(
                "Gotten\n\n%s\n\nExpected\n\n%s\n\n", got, "\n".join(expected)
            )
            raise

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )
    def test_ReportBuildSequencePlain(self):
        _logger.info("#" * 100)
        self.runTestBuildSequenceTable(tablefmt="plain")

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )  # pylint: disable=invalid-name
    def test_ReportBuildSequenceFallback(self):
        _logger.info("#" * 100)
        with patch.object(self.server, "client_capabilities", None):
            self.runTestBuildSequenceTable(tablefmt="plain")

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )  # pylint: disable=invalid-name
    def test_ReportBuildSequenceMarkdown(self):
        _logger.info("#" * 100)
        with patch.object(
            self.server,
            "client_capabilities",
            ClientCapabilities(
                text_document=TextDocumentClientCapabilities(
                    synchronization=None,
                    completion=None,
                    hover=HoverAbstract(
                        dynamic_registration=False,
                        # This is what we really need
                        content_format=[MarkupKind.Markdown,],
                    ),
                    signature_help=None,
                    references=None,
                    document_highlight=None,
                    document_symbol=None,
                    formatting=None,
                    range_formatting=None,
                    on_type_formatting=None,
                    definition=None,
                    type_definition=None,
                    implementation=None,
                    code_action=None,
                    code_lens=None,
                    document_link=None,
                    color_provider=None,
                    rename=None,
                    publish_diagnostics=PublishDiagnosticsAbstract(
                        related_information=True
                    ),
                    folding_range=None,
                )
            ),
        ):
            self.runTestBuildSequenceTable(tablefmt="github")

    @patch.object(
        hdl_checker.core.HdlCheckerCore,
        "resolveDependencyToPath",
        lambda self, _: None,
    )  # pylint: disable=invalid-name
    def test_DependencyInfoForPathNotFound(self):  # pylint: disable=invalid-name
        _logger.info("#" * 100)
        path = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))
        dependency = RequiredDesignUnit(
            name=Identifier("clock_divider"),
            library=Identifier("basic_library"),
            owner=path,
            locations=(),
        )
        self.assertEqual(
            self.server.getDependencyInfoForHover(dependency),
            "Couldn't find a source defining 'basic_library.clock_divider'",
        )

    @patch.object(
        hdl_checker.core.HdlCheckerCore,
        "resolveDependencyToPath",
        lambda self, _: (Path("some_path"), Identifier("some_library")),
    )
    def test_ReportDependencyInfo(self):
        _logger.info("#" * 100)
        path = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))
        dependency = RequiredDesignUnit(
            name=Identifier("clock_divider"),
            library=Identifier("basic_library"),
            owner=path,
            locations=(),
        )
        self.assertEqual(
            self.server.getDependencyInfoForHover(dependency),
            'Path "some_path", library "some_library"',
        )

    def test_ReportDesignUnitAccordingToPosition(  # pylint: disable=invalid-name
        self,
    ) -> None:
        _logger.info("#" * 100)
        unit_a = VhdlDesignUnit(
            owner=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
            type_=DesignUnitType.entity,
            name="unit_a",
            locations=(Location(line=1, column=2), Location(line=3, column=4)),
        )

        unit_b = VerilogDesignUnit(
            owner=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
            type_=DesignUnitType.package,
            name="unit_b",
            locations=(Location(line=5, column=6), Location(line=7, column=8)),
        )

        dep_a = RequiredDesignUnit(
            name=Identifier("dep_a"),
            library=Identifier("lib_a"),
            owner=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
            locations=(Location(line=9, column=10), Location(line=11, column=12)),
        )

        dep_b = RequiredDesignUnit(
            name=Identifier("dep_a"),
            library=Identifier("lib_a"),
            owner=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
            locations=(Location(line=13, column=14), Location(line=15, column=16)),
        )

        def getDesignUnitsByPath(self, path):  # pylint: disable=unused-argument
            if path != Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")):
                self.fail("Expected foo.vhd but got %s" % path)
            return {unit_a, unit_b}

        def getDependenciesByPath(self, path):  # pylint: disable=unused-argument
            if path != Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")):
                self.fail("Expected foo.vhd but got %s" % path)
            return {dep_a, dep_b}

        patches = (
            patch.object(
                hdl_checker.database.Database,
                "getDesignUnitsByPath",
                getDesignUnitsByPath,
            ),
            patch.object(
                hdl_checker.database.Database,
                "getDependenciesByPath",
                getDependenciesByPath,
            ),
        )

        path = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))

        for _patch in patches:
            _patch.start()

        # Check locations outside return nothing
        self.assertIsNone(self.server.getElementAtPosition(path, Position(0, 0)))

        # Check design units are found, ensure boundaries match
        self.assertIsNone(self.server.getElementAtPosition(path, Position(1, 1)))
        self.assertIs(self.server.getElementAtPosition(path, Position(1, 2)), unit_a)
        self.assertIs(self.server.getElementAtPosition(path, Position(1, 7)), unit_a)
        self.assertIsNone(self.server.getElementAtPosition(path, Position(1, 8)))

        self.assertIsNone(self.server.getElementAtPosition(path, Position(3, 3)))
        self.assertIs(self.server.getElementAtPosition(path, Position(3, 4)), unit_a)
        self.assertIs(self.server.getElementAtPosition(path, Position(3, 9)), unit_a)
        self.assertIsNone(self.server.getElementAtPosition(path, Position(3, 10)))

        self.assertIsNone(self.server.getElementAtPosition(path, Position(5, 5)))
        self.assertIs(self.server.getElementAtPosition(path, Position(5, 6)), unit_b)
        self.assertIs(self.server.getElementAtPosition(path, Position(5, 11)), unit_b)
        self.assertIsNone(self.server.getElementAtPosition(path, Position(5, 12)))

        self.assertIsNone(self.server.getElementAtPosition(path, Position(7, 7)))
        self.assertIs(self.server.getElementAtPosition(path, Position(7, 8)), unit_b)
        self.assertIs(self.server.getElementAtPosition(path, Position(7, 13)), unit_b)
        self.assertIsNone(self.server.getElementAtPosition(path, Position(7, 14)))

        # Now check dependencies
        self.assertIsNone(self.server.getElementAtPosition(path, Position(9, 9)))
        self.assertIs(self.server.getElementAtPosition(path, Position(9, 10)), dep_a)
        self.assertIs(self.server.getElementAtPosition(path, Position(9, 20)), dep_a)
        self.assertIsNone(self.server.getElementAtPosition(path, Position(9, 21)))

        self.assertIsNone(self.server.getElementAtPosition(path, Position(11, 11)))
        self.assertIs(self.server.getElementAtPosition(path, Position(11, 12)), dep_a)
        self.assertIs(self.server.getElementAtPosition(path, Position(11, 22)), dep_a)
        self.assertIsNone(self.server.getElementAtPosition(path, Position(11, 23)))

        self.assertIsNone(self.server.getElementAtPosition(path, Position(13, 13)))
        self.assertIs(self.server.getElementAtPosition(path, Position(13, 14)), dep_b)
        self.assertIs(self.server.getElementAtPosition(path, Position(13, 24)), dep_b)
        self.assertIsNone(self.server.getElementAtPosition(path, Position(13, 25)))

        self.assertIsNone(self.server.getElementAtPosition(path, Position(15, 15)))
        self.assertIs(self.server.getElementAtPosition(path, Position(15, 16)), dep_b)
        self.assertIs(self.server.getElementAtPosition(path, Position(15, 26)), dep_b)
        self.assertIsNone(self.server.getElementAtPosition(path, Position(15, 27)))

        for _patch in patches:
            _patch.stop()

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )
    def test_HoverOnInvalidRange(self):
        _logger.info("#" * 100)
        self.assertIsNone(
            self.client.lsp.send_request(
                features.HOVER,
                HoverParams(
                    TextDocumentIdentifier(
                        uris.from_fs_path(
                            p.join(TEST_PROJECT, "another_library", "foo.vhd")
                        )
                    ),
                    Position(line=0, character=0),
                ),
            ).result(LSP_REQUEST_TIMEOUT)
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )
    def test_HoverOnDesignUnit(self):
        _logger.info("#" * 100)
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")
        very_common_pkg = p.join(TEST_PROJECT, "basic_library", "very_common_pkg.vhd")
        package_with_constants = p.join(
            TEST_PROJECT, "basic_library", "package_with_constants.vhd"
        )
        clock_divider = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")

        expected = [
            "Build sequence for %s is" % str(path_to_foo),
            "",
            tabulate(
                [
                    (1, "basic_library", str(very_common_pkg)),
                    (2, "basic_library", str(package_with_constants)),
                    (3, "basic_library", str(clock_divider)),
                    (4, DEFAULT_LIBRARY.name, str(path_to_foo)),
                ],
                headers=("#", "Library", "Path"),
                tablefmt="plain",
            ),
        ]

        self.assertEqual(
            self.client.lsp.send_request(
                features.HOVER,
                HoverParams(
                    TextDocumentIdentifier(uris.from_fs_path(path_to_foo)),
                    Position(line=7, character=7),
                ),
            )
            .result(LSP_REQUEST_TIMEOUT)
            .contents,
            "\n".join(expected),
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )
    def test_HoverOnDependency(self):
        _logger.info("#" * 100)
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")
        clock_divider = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")

        self.assertEqual(
            self.client.lsp.send_request(
                features.HOVER,
                HoverParams(
                    TextDocumentIdentifier(uris.from_fs_path(path_to_foo)),
                    Position(line=32, character=32),
                ),
            )
            .result(LSP_REQUEST_TIMEOUT)
            .contents,
            'Path "%s", library "basic_library"' % clock_divider,
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )  # pylint: disable=invalid-name
    def test_GetDefinitionMatchingDependency(self):
        _logger.info("#" * 100)
        source = p.join(TEST_PROJECT, "basic_library", "use_entity_a_and_b.vhd")
        target = p.join(TEST_PROJECT, "basic_library", "two_entities_one_file.vhd")

        definitions = {
            (
                x.uri,
                x.range.start.line,
                x.range.start.character,
                x.range.end.line,
                x.range.end.character,
            )
            for x in self.client.lsp.send_request(
                features.DEFINITION,
                TextDocumentPositionParams(
                    TextDocumentIdentifier(uris.from_fs_path(source)), Position(1, 9),
                ),
            ).result(LSP_REQUEST_TIMEOUT)
        }

        self.assertIn(
            (uris.from_fs_path(target), 1, 7, 1, 15), definitions,
        )

        self.assertIn(
            (uris.from_fs_path(target), 4, 7, 4, 15), definitions,
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )  # pylint: disable=invalid-name
    def test_GetDefinitionBuiltInLibrary(self) -> None:
        _logger.info("#" * 100)
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")

        self.assertFalse(
            self.server.definitions(
                TextDocumentPositionParams(
                    TextDocumentIdentifier(uris.from_fs_path(path_to_foo)),
                    Position(3, 15),
                )
            )
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        (Identifier("ieee"),),
    )
    def test_GetDefinitionNotKnown(self) -> None:
        _logger.info("#" * 100)
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")

        self.assertFalse(
            self.server.definitions(
                TextDocumentPositionParams(
                    TextDocumentIdentifier(uris.from_fs_path(path_to_foo)),
                    Position(0, 0),
                )
            )
        )

    @patch.object(
        hdl_checker.database.Database,
        "getReferencesToDesignUnit",
        return_value=[
            RequiredDesignUnit(
                name=Identifier("clock_divider"),
                library=Identifier("basic_library"),
                owner=Path("some_path"),
                locations=(Location(1, 2), Location(3, 4)),
            )
        ],
    )
    def test_ReferencesOfAValidElement(self, get_references) -> Any:
        _logger.info("#" * 100)
        # We'll pass the path to foo.vhd but we're patching the
        # getReferencesToDesignUnit to return "some_path"
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")

        # Make sure we picked up an existing element
        unit = self.server.getElementAtPosition(Path(path_to_foo), Position(7, 7))
        self.assertIsNotNone(unit)

        references = [
            (
                x.uri,
                x.range.start.line,
                x.range.start.character,
                x.range.end.line,
                x.range.end.character,
            )
            for x in self.client.lsp.send_request(
                features.REFERENCES,
                ReferenceParams(
                    TextDocumentIdentifier(uris.from_fs_path(path_to_foo)),
                    Position(7, 7),
                    ReferenceContext(include_declaration=False),
                ),
            ).result(LSP_REQUEST_TIMEOUT)
        ]

        self.assertCountEqual(
            references,
            {
                (uris.from_fs_path("some_path"), 1, 2, 1, 3),
                (uris.from_fs_path("some_path"), 3, 4, 3, 5),
            },
        )

        get_references.assert_called_once()
        get_references.reset_mock()

        references = [
            (
                x.uri,
                x.range.start.line,
                x.range.start.character,
                x.range.end.line,
                x.range.end.character,
            )
            for x in self.client.lsp.send_request(
                features.REFERENCES,
                ReferenceParams(
                    TextDocumentIdentifier(uris.from_fs_path(path_to_foo)),
                    Position(7, 7),
                    ReferenceContext(include_declaration=True),
                ),
            ).result(LSP_REQUEST_TIMEOUT)
        ]

        self.assertCountEqual(
            references,
            {
                (uris.from_fs_path(path_to_foo), 7, 7, 7, 8),
                (uris.from_fs_path("some_path"), 1, 2, 1, 3),
                (uris.from_fs_path("some_path"), 3, 4, 3, 5),
            },
        )

    def test_ReferencesOfAnInvalidElement(self):  # pylint: disable=invalid-name
        _logger.info("#" * 100)
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")

        # Make sure there's no element at this location
        unit = self.server.getElementAtPosition(Path(path_to_foo), Position(0, 0))
        self.assertIsNone(unit)

        for include_declaration in (False, True):
            self.assertIsNone(
                self.server.references(
                    ReferenceParams(
                        TextDocumentIdentifier(uris.from_fs_path(path_to_foo)),
                        Position(0, 0),
                        ReferenceContext(include_declaration=include_declaration),
                    )
                )
            )

    def test_changeConfiguration(self):
        _logger.info("#" * 100)
        # pylint: disable=no-member
        with patch.object(self.server, "onConfigUpdate"):
            self.client.lsp.send_request(
                features.WORKSPACE_DID_CHANGE_CONFIGURATION, {"foo": "bar"}
            ).result(LSP_REQUEST_TIMEOUT)

            self.assertIn(
                "bar", {x[0].foo for x in self.server.onConfigUpdate.call_args if x},
            )
        # pylint: enable=no-member



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined, missing-docstring, protected-access

import json
import logging
from tempfile import NamedTemporaryFile

from parameterized import parameterized_class  # type: ignore

from hdl_checker.tests import TestCase, writeListToFile

from hdl_checker.parsers.elements.dependency_spec import (
    RequiredDesignUnit,
    IncludedPath,
)
from hdl_checker.parsers.elements.identifier import VerilogIdentifier
from hdl_checker.parsers.elements.parsed_element import Location
from hdl_checker.parsers.verilog_parser import VerilogDesignUnit, VerilogParser
from hdl_checker.path import Path
from hdl_checker.serialization import StateEncoder, jsonObjectHook
from hdl_checker.types import DesignUnitType

_logger = logging.getLogger(__name__)


def parametrizeClassWithFileTypes(cls):
    keys = ["filetype"]
    values = [(x,) for x in ("v", "vh", "sv", "svh")]

    return parameterized_class(keys, values)(cls)


@parametrizeClassWithFileTypes
class TestVerilogSource(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.filename = NamedTemporaryFile(suffix="." + cls.filetype).name

        writeListToFile(
            cls.filename,
            [
                '`include "some/include"',
                "",
                "import some_package::*;",
                "import  another_package :: some_name ;",
                "",
                "module clock_divider",
                "    #(parameter DIVISION = 5)",
                "    (// Usual ports",
                "    input clk,",
                "    input rst,",
                "    // Output clock divided",
                "    output       clk_div);",
                "   localparam foo::bar = std::randomize(cycles);",
                "endmodule",
                "",
                "package \\m$gPkg! ;",
                "  integer  errCnt  = 0;",
                "  integer  warnCnt = 0;",
                "endpackage",
                "",
            ],
        )

        cls.source = VerilogParser(Path(cls.filename))

    def test_GetDesignUnits(self):
        design_units = list(self.source.getDesignUnits())
        _logger.debug("Design units: %s", design_units)
        self.assertCountEqual(
            design_units,
            [
                VerilogDesignUnit(
                    owner=self.source.filename,
                    name="clock_divider",
                    type_=DesignUnitType.entity,
                    locations={(5, 7)},
                ),
                VerilogDesignUnit(
                    owner=self.source.filename,
                    name="\\m$gPkg!",
                    type_=DesignUnitType.package,
                    locations={(15, 8)},
                ),
            ],
        )

    def test_GetDependencies(self):
        expected = [
            IncludedPath(
                owner=Path(self.filename),
                name=VerilogIdentifier("some/include"),
                locations=(Location(line=0, column=8),),
            )
        ]

        if self.filetype in ("sv", "svh"):
            expected += [
                RequiredDesignUnit(
                    owner=Path(self.filename),
                    name=VerilogIdentifier("some_package"),
                    library=None,
                    locations=(Location(line=2, column=7),),
                ),
                RequiredDesignUnit(
                    owner=Path(self.filename),
                    name=VerilogIdentifier("another_package"),
                    library=None,
                    locations=(Location(line=3, column=8),),
                ),
                RequiredDesignUnit(
                    owner=Path(self.filename),
                    name=VerilogIdentifier("foo"),
                    library=None,
                    locations=(Location(line=12, column=14),),
                ),
            ]

        self.assertCountEqual(self.source.getDependencies(), expected)

    def test_CacheRecovery(self):
        state = json.dumps(self.source, cls=StateEncoder)
        _logger.info("State before: %s", state)
        recovered = json.loads(state, object_hook=jsonObjectHook)
        self.assertEqual(self.source.filename, recovered.filename)



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Language server protocol implementation"

import json
import logging
from os import getpid
from os import path as p
from tempfile import mkdtemp
from typing import Any, Callable, Iterable, List, Optional, Set, Tuple, Union

from pygls.features import (
    DEFINITION,
    HOVER,
    INITIALIZE,
    INITIALIZED,
    REFERENCES,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    WORKSPACE_DID_CHANGE_CONFIGURATION,
)
from pygls.server import LanguageServer
from pygls.types import (
    ClientCapabilities,
    Diagnostic,
    DiagnosticSeverity,
    DidChangeConfigurationParams,
    DidChangeTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    Hover,
    HoverParams,
    InitializeParams,
    Location,
    MarkupKind,
    MessageType,
    Position,
    Range,
    ReferenceParams,
    TextDocumentPositionParams,
)
from pygls.uris import from_fs_path, to_fs_path
from tabulate import tabulate

from . import DEFAULT_LIBRARY, DEFAULT_PROJECT_FILE
from .config_generators.simple_finder import SimpleFinder
from .core import HdlCheckerCore
from .diagnostics import CheckerDiagnostic, DiagType
from .exceptions import UnknownParameterError
from .parsers.elements.dependency_spec import (
    BaseDependencySpec,
    IncludedPath,
    RequiredDesignUnit,
)
from .parsers.elements.design_unit import (
    VerilogDesignUnit,
    VhdlDesignUnit,
    tAnyDesignUnit,
)
from .path import Path, TemporaryPath
from .types import ConfigFileOrigin  # , Location
from .utils import debounce, getTemporaryFilename, logCalls, onNewReleaseFound

_logger = logging.getLogger(__name__)

AUTO_PROJECT_FILE_NAME = "project.json"
LINT_DEBOUNCE_S = 0.5  # 500 ms

URI = str


def _translateSeverity(severity: DiagType) -> DiagnosticSeverity:
    """
    Translate hdl_checker's DiagType into pygls's DiagnosticSeverity into LSP
    severity
    """
    if severity in (
        DiagType.STYLE_WARNING,
        DiagType.STYLE_ERROR,
        DiagType.INFO,
        DiagType.STYLE_INFO,
    ):
        return DiagnosticSeverity.Information
    if severity in (DiagType.WARNING,):
        return DiagnosticSeverity.Warning
    if severity in (DiagType.ERROR,):
        return DiagnosticSeverity.Error
    return DiagnosticSeverity.Error


def checkerDiagToLspDict(diag: CheckerDiagnostic) -> Diagnostic:
    """
    Converts a CheckerDiagnostic object into pygls.Diagnostic type expected by
    the publish_diagnostics LSP method
    """
    _logger.debug(diag)
    return Diagnostic(
        range=Range(
            start=Position(
                line=diag.line_number or 0, character=diag.column_number or 0
            ),
            end=Position(
                line=diag.line_number or 0, character=(diag.column_number or 0) + 1
            ),
        ),
        message=diag.text,
        severity=_translateSeverity(diag.severity),
        code=diag.error_code if diag.error_code else None,
        source=diag.checker,
    )


class Server(HdlCheckerCore):
    """
    HDL Checker project builder class
    """

    def __init__(self, lsp, root_dir):
        # type: (LanguageServer, Path) -> None
        self._lsp = lsp
        super().__init__(root_dir)

    def _handleUiInfo(self, message):
        # type: (...) -> Any
        _logger.debug("UI info: %s (lsp=%s)", message, self._lsp)
        if self._lsp:  # pragma: no cover
            self._lsp.show_message(message, MessageType.Info)

    def _handleUiWarning(self, message):
        # type: (...) -> Any
        _logger.debug("UI warning: %s (lsp=%s)", message, self._lsp)
        if self._lsp:  # pragma: no cover
            self._lsp.show_message(message, MessageType.Warning)

    def _handleUiError(self, message):
        # type: (...) -> Any
        _logger.debug("UI error: %s (lsp=%s)", message, self._lsp)
        if self._lsp:  # pragma: no cover
            self._lsp.show_message(message, MessageType.Error)


class HdlCheckerLanguageServer(LanguageServer):
    """
    Implementation of the Microsoft VSCode Language Server Protocol
    https://github.com/Microsoft/language-server-protocol/blob/master/versions/protocol-1-x.md
    """

    def __init__(self, *args, **kwargs) -> None:
        self._checker: Optional[Server] = None
        super().__init__(*args, **kwargs)
        # Default checker
        self.onConfigUpdate(None)
        self._global_diags: Set[CheckerDiagnostic] = set()
        self.initialization_options: Optional[Any] = None
        self.client_capabilities: Optional[ClientCapabilities] = None

    @property
    def checker(self) -> Server:
        """
        Returns a valid checker, either the one configured during
        HdlCheckerLanguageServer.onConfigUpdate or a new one using a temporary
        directory.
        """
        if self._checker is None:
            _logger.debug("Server was not initialized, using a temporary one")
            root_dir = mkdtemp(prefix="temp_hdl_checker_pid{}_".format(getpid()))
            self._checker = Server(self, root_dir=TemporaryPath(root_dir))
            _logger.debug("Temporary server %s is ready", self._checker)
        return self._checker

    def showInfo(self, msg: str) -> None:
        """
        Shorthand for self.show_message(msg, MessageType.Info)
        """
        _logger.info("[INFO] %s", msg)
        self.show_message(msg, MessageType.Info)

    def showWarning(self, msg: str) -> None:
        """
        Shorthand for self.show_message(msg, MessageType.Warning)
        """
        _logger.info("[WARNING] %s", msg)
        self.show_message(msg, MessageType.Warning)

    def onConfigUpdate(self, options: Optional[Any]) -> None:
        """
        Updates the checker server from options if the 'project_file' key is
        present. Please not that this is run from both initialize and
        workspace/did_change_configuration and when ran initialize the LSP
        client might not ready to take messages. To circumvent this, make sure
        m_initialize returns before calling this to actually configure the
        server.
        """
        if not self.workspace or not self.workspace.root_uri:
            return

        root_dir = to_fs_path(self.workspace.root_uri)
        self._checker = Server(self, root_dir=Path(root_dir))

        _logger.debug("Updating from %s, workspace=%s", options, self.workspace)

        # Clear previus diagnostics
        self._global_diags = set()

        path = self._getProjectFilePath(options)

        try:
            self.checker.setConfig(path, origin=ConfigFileOrigin.user)
            return
        except UnknownParameterError as exc:
            _logger.info("Failed to read config from %s: %s", path, exc)
            return
        except FileNotFoundError:
            # If the file couldn't be found, proceed to searching the root
            # URI (if it has been set)
            pass

        if not self.workspace or not self.workspace.root_path:
            _logger.debug("No workspace and/or root path not set, can't search files")
            return

        self.showInfo("Searching {} for HDL files...".format(self.workspace.root_path))

        # Having no project file but with root URI triggers searching for
        # sources automatically
        config = SimpleFinder([self.workspace.root_path]).generate()

        # Write this to a file and tell the server to use it
        auto_project_file = getTemporaryFilename(AUTO_PROJECT_FILE_NAME)
        with open(auto_project_file, "w") as fd:
            json.dump(config, fd)
        self.checker.setConfig(auto_project_file, origin=ConfigFileOrigin.generated)

    def _getProjectFilePath(self, options: Optional[Any] = None) -> str:
        """
        Tries to get 'project_file' from the options dict and combine it with
        the root URI as provided by the workspace
        """
        path = DEFAULT_PROJECT_FILE
        if options and options.project_file is not None:
            path = options.project_file

        # Project file will be related to the root path
        if self.workspace:
            path = p.join(self.workspace.root_path, path)

        return path

    @debounce(LINT_DEBOUNCE_S, keyed_by="uri")
    def lint(self, uri: URI, is_saved: bool) -> None:
        """
        Check a file for lint errors
        """
        _logger.debug("Linting %s (file was %s saved)", uri, "" if is_saved else "not")
        diags = set(self._getDiags(uri, is_saved))

        # Separate the diagnostics in filename groups to publish diagnostics
        # referring to all paths
        paths = {diag.filename for diag in diags}
        # Add text_doc.uri to the set to trigger clearing diagnostics when it's not
        # present
        paths.add(Path(to_fs_path(uri)))

        for path in paths:
            self.lsp.publish_diagnostics(
                from_fs_path(str(path)),
                tuple(
                    checkerDiagToLspDict(diag)
                    for diag in diags
                    if diag.filename == path
                ),
            )

    def _getDiags(self, doc_uri: URI, is_saved: bool) -> Iterable[CheckerDiagnostic]:
        """
        Gets diags of the URI, wether from the saved file or from its contents;
        returns an iterable containing the diagnostics of the doc_uri and other
        URIs that were compiled as dependencies and generated diagnostics with
        severity higher than error
        """
        if self.checker is None:  # pragma: no cover
            _logger.debug("No checker, won't try to get diagnostics")
            return ()

        # If the file has not been saved, use the appropriate method, which
        # will involve dumping the modified contents into a temporary file
        path = Path(to_fs_path(doc_uri))

        if is_saved:
            return self.checker.getMessagesByPath(path)
        text = self.workspace.get_document(doc_uri).source
        return self.checker.getMessagesWithText(path, text)

    def references(self, params: ReferenceParams) -> Optional[List[Location]]:
        "Tries to find references for the selected element"

        element = self.getElementAtPosition(
            Path(to_fs_path(params.textDocument.uri)), params.position
        )

        # Element not identified
        if element is None:
            return None

        references: List[Location] = []

        if params.context.includeDeclaration:
            for line, column in element.locations:
                references += [
                    Location(
                        uri=from_fs_path(str(element.owner)),
                        range=Range(
                            start=Position(line, column), end=Position(line, column + 1)
                        ),
                    )
                ]

        for reference in self.checker.database.getReferencesToDesignUnit(element):
            for line, column in reference.locations:
                references += [
                    Location(
                        uri=from_fs_path(str(reference.owner)),
                        range=Range(
                            start=Position(line, column), end=Position(line, column + 1)
                        ),
                    )
                ]

        return references

    @property
    def _use_markdown_for_hover(self):
        """
        Returns True if the client has reported 'markdown' as one of the
        supported formats, i.e., 'markdown' is present inside
        TextDocumentClientCapabilities.hover.contentFormat
        """
        try:
            return (
                MarkupKind.Markdown.value
                in self.client_capabilities.textDocument.hover.contentFormat
            )
        except AttributeError:
            return False

    def _format(self, text):
        """
        Double line breaks if workspace supports markdown
        """
        if self._use_markdown_for_hover:
            return text.replace("\n", "\n\n")

        return text

    def getBuildSequenceForHover(self, path: Path) -> str:
        """
        Return a formatted text with the build sequence for the given path
        """
        sequence = []  # type: List[Tuple[int, str, str]]

        # Adds the sequence of dependencies' paths
        for i, (seq_library, seq_path) in enumerate(
            self.checker.database.getBuildSequence(
                path, self.checker.builder.builtin_libraries
            ),
            1,
        ):
            sequence += [(i, str(seq_library), str(seq_path))]

        # Adds the original path
        sequence += [
            (
                len(sequence) + 1,
                str(self.checker.database.getLibrary(path) or DEFAULT_LIBRARY),
                str(path),
            )
        ]

        return "Build sequence for {} is\n\n{}".format(
            path,
            tabulate(
                sequence,
                tablefmt="github" if self._use_markdown_for_hover else "plain",
                headers=("#", "Library", "Path"),
            ),
        )

    def getDependencyInfoForHover(self, dependency):
        # type: (BaseDependencySpec) -> str
        """
        Report which source defines a given dependency when the user hovers
        over its name
        """
        if isinstance(dependency, RequiredDesignUnit):
            info = self.checker.resolveDependencyToPath(dependency)
            if info is not None:
                return self._format('Path "{}", library "{}"'.format(info[0], info[1]))
            return "Couldn't find a source defining '{}.{}'".format(
                dependency.library, dependency.name
            )

        if isinstance(dependency, IncludedPath):
            info = self.checker.database.resolveIncludedPath(dependency)
            if info is not None:
                return self._format('Path "{}"'.format(info))
        return "Couldn't find a source defining '{}'".format(dependency.name)

    def getElementAtPosition(
        self, path: Path, position: Position
    ) -> Union[BaseDependencySpec, tAnyDesignUnit, None]:
        """
        Gets design units and dependencies (in this order) of path and checks
        if their definitions include position. Not every element is identified,
        only those pertinent to the core functionality, e.g. design units and
        dependencies.
        """
        for meth in (
            self.checker.database.getDesignUnitsByPath,
            self.checker.database.getDependenciesByPath,
        ):  # type: Callable
            for element in meth(path):
                if element.includes(position.line, position.character):
                    return element

        return None

    def hover(self, params: HoverParams) -> Optional[Hover]:
        """
        Handles HoverParams and produces a Hover object if a known element is
        found within the given location
        """
        path = Path(to_fs_path(params.textDocument.uri))
        # Check if the element under the cursor matches something we know
        element = self.getElementAtPosition(path, params.position)

        _logger.debug("Getting info from %s", element)

        if not isinstance(
            element, (VerilogDesignUnit, VhdlDesignUnit, BaseDependencySpec)
        ):
            return None

        if isinstance(element, (VerilogDesignUnit, VhdlDesignUnit)):
            contents = self.getBuildSequenceForHover(path)
        else:
            contents = self.getDependencyInfoForHover(element)

        return Hover(
            contents=contents,
            range=Range(
                start=Position(
                    line=params.position.line, character=params.position.character
                ),
                end=Position(
                    line=params.position.line, character=params.position.character + 1
                ),
            ),
        )

    @logCalls
    def definitions(
        self, params: TextDocumentPositionParams
    ) -> Optional[List[Location]]:
        """
        Returns known definitions found in the given location
        """
        dependency = self.getElementAtPosition(
            Path(to_fs_path(params.textDocument.uri)), params.position
        )

        if not isinstance(dependency, BaseDependencySpec):
            _logger.debug("Go to definition not supported for item %s", dependency)
            return []

        # Work out where this dependency refers to
        info = self.checker.resolveDependency(dependency)

        if info is None:
            _logger.debug("Unable to resolve %s to a path", dependency)
            return []

        _logger.info("Dependency %s resolved to %s", dependency, info)

        # Make the response
        target_path, _ = info
        target_uri = from_fs_path(str(target_path))

        if not target_path:
            _logger.debug("Unable to resolve %s to a path", dependency)
            return []

        # Included paths are dependencies but they're referred to by path, so
        # we return a definition to point to the beginning of the file
        if isinstance(dependency, IncludedPath):
            return [Location(target_uri, Range(Position(0, 0), Position(0, 1)))]

        locations: List[Location] = []

        # Get the design unit that has matched the dependency to extract the
        # location where it's defined
        for unit in self.checker.database.getDesignUnitsByPath(target_path):
            if unit.name == dependency.name and unit.locations:
                for line, column in unit.locations:
                    locations += [
                        Location(
                            target_uri,
                            Range(
                                Position(line, column),
                                Position(line, column + len(unit)),
                            ),
                        )
                    ]

        return locations


def setupLanguageServerFeatures(server: HdlCheckerLanguageServer) -> None:
    """Adds pygls features to an instance of HdlCheckerLanguageServer"""

    # pylint: disable=unused-variable
    @server.feature(INITIALIZE)
    def initialize(self: HdlCheckerLanguageServer, params: InitializeParams) -> None:
        options = params.initializationOptions
        self.client_capabilities = params.capabilities
        self.initialization_options = options

    @server.feature(INITIALIZED)
    def initialized(self: HdlCheckerLanguageServer, *_):
        """
        Enables processing of actions that were generated upon m_initialize and
        were delayed because the client might need further info (for example to
        handle window/showMessage requests)
        """
        self.onConfigUpdate(self.initialization_options)
        onNewReleaseFound(self.showInfo)

    @server.feature(TEXT_DOCUMENT_DID_SAVE)
    def didSave(self: HdlCheckerLanguageServer, params: DidSaveTextDocumentParams):
        """Text document did change notification."""
        self.lint(params.textDocument.uri, True)

    @server.feature(TEXT_DOCUMENT_DID_CHANGE)
    def didChange(self: HdlCheckerLanguageServer, params: DidChangeTextDocumentParams):
        """Text document did change notification."""
        self.lint(params.textDocument.uri, False)

    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    def didOpen(self: HdlCheckerLanguageServer, params: DidOpenTextDocumentParams):
        """Text document did change notification."""
        self.lint(params.textDocument.uri, True)

    @server.feature(WORKSPACE_DID_CHANGE_CONFIGURATION)
    def didChangeConfiguration(
        self: HdlCheckerLanguageServer, settings: DidChangeConfigurationParams = None
    ) -> None:
        self.onConfigUpdate(settings)

    @server.feature(HOVER)
    def onHover(self: HdlCheckerLanguageServer, params: HoverParams) -> Optional[Hover]:
        return self.hover(params)

    @server.feature(REFERENCES)
    def onReferences(
        self: HdlCheckerLanguageServer, params: ReferenceParams
    ) -> Optional[List[Location]]:
        return self.references(params)

    @server.feature(DEFINITION)
    def onDefinition(
        self: HdlCheckerLanguageServer, params: TextDocumentPositionParams
    ) -> Optional[List[Location]]:
        return self.definitions(params)

    # pylint: enable=unused-variable



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined, missing-docstring, protected-access

import asyncio
import logging
import os
import os.path as p
import subprocess as subp
import tempfile
import time
from multiprocessing import Event, Process, Queue
from threading import Thread

import requests
from mock import patch
from pygls import features, uris
from pygls.types import ClientCapabilities, Diagnostic, InitializeParams

from nose2.tools import such  # type: ignore

from hdl_checker.tests import disableVunit, getTestTempPath

import hdl_checker.lsp
from hdl_checker import server
from hdl_checker.utils import ON_LINUX, ON_WINDOWS, isProcessRunning, terminateProcess

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_LOG_PATH = p.join(os.environ["TOX_ENV_DIR"], "log")
SERVER_LOG_LEVEL = os.environ.get("SERVER_LOG_LEVEL", "WARNING")

HDL_CHECKER_BASE_PATH = p.abspath(p.join(p.dirname(__file__), "..", ".."))

CALL_TIMEOUT = 5

# Static shutdown message to avoid havint to create a server/client pair
LSP_SHUTDOWN = b"\r\n".join(
    [
        b"Content-Length: 102",
        b"Content-Type: application/vscode-jsonrpc; charset=utf-8",
        b"",
        b'{"id": "e7389d0d-4d3c-432d-8506-68a1b2faca48", "jsonrpc": "2.0", "met     hod": "shutdown", "params": null}',
    ]
)


def _path(*args):
    # type: (str) -> str
    "Helper to reduce foorprint of p.join(TEST_TEMP_PATH, *args)"
    return p.join(TEST_TEMP_PATH, *args)


def doNothing(queue):
    _logger.debug("I'm ready")
    queue.get()
    _logger.debug("Ok, done")


def _getUnusedLocalhostPort():
    """
    These were "Borrowed" from YCM.
    See https://github.com/Valloric/YouCompleteMe
    """
    import socket

    sock = socket.socket()
    # This tells the OS to give us any free port in the range [1024 - 65535]
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def _startClient(client):
    client.start()


class _ClientServer(
    object
):  # pylint: disable=useless-object-inheritance,too-few-public-methods
    """ A class to setup a client/server pair """

    def __init__(self):
        #  # Client to Server pipe
        #  csr, csw = os.pipe()
        #  # Server to client pipe
        #  scr, scw = os.pipe()

        #  self.server_thread = Thread(
        #      target=start_io_lang_server,
        #      args=(
        #          os.fdopen(csr, "rb"),
        #          os.fdopen(scw, "wb"),
        #          False,
        #          hdl_checker.lsp.HdlCheckerLanguageServer,
        #      ),
        #  )

        #  self.server_thread.daemon = True
        #  self.server_thread.start()

        #  # Object being tested is the server thread. Avoid both objects
        #  # competing for the same cache by using the raw Python language server
        #  self.client = PythonLanguageServer(
        #      os.fdopen(scr, "rb"), os.fdopen(csw, "wb"), start_io_lang_server
        #  )

        #  self.client_thread = Thread(target=_startClient, args=[self.client])
        #  self.client_thread.daemon = True
        #  self.client_thread.start()

        # Client to Server pipe
        csr, csw = os.pipe()
        # Server to client pipe
        scr, scw = os.pipe()

        self.server = hdl_checker.lsp.HdlCheckerLanguageServer()
        hdl_checker.lsp.setupLanguageServerFeatures(self.server)
        self.server.show_message = lambda *args, **kwargs: _logger.fatal(
            "%s, %s", args, kwargs
        )

        server_thread = Thread(
            target=self.server.start_io,
            args=(os.fdopen(csr, "rb"), os.fdopen(scw, "wb")),
        )

        server_thread.daemon = True
        server_thread.start()

        # Add thread id to the server (just for testing)
        self.server.thread_id = server_thread.ident

        # Setup client
        self.client = hdl_checker.lsp.HdlCheckerLanguageServer(asyncio.new_event_loop())
        self.client_diagnostics = []

        @self.client.feature(features.TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)
        def dbg_publish(diag: Diagnostic):
            _logger.info("Client received diagnostic: %s", diag)
            self.client_diagnostics.append(diag)

        client_thread = Thread(
            target=self.client.start_io,
            args=(os.fdopen(scr, "rb"), os.fdopen(csw, "wb")),
        )

        client_thread.daemon = True
        client_thread.start()


such.unittest.TestCase.maxDiff = None

with such.A("hdl_checker server") as it:

    _SERVER_BASE_CMD = [
        "coverage",
        "run",
        p.join(HDL_CHECKER_BASE_PATH, "hdl_checker", "server.py"),
        "--log-level",
        SERVER_LOG_LEVEL,
        "--stdout",
        p.join(TEST_LOG_PATH, "hdl_checker-stdout.log"),
        "--stderr",
        p.join(TEST_LOG_PATH, "hdl_checker-stderr.log"),
        "--log-stream",
        p.join(TEST_LOG_PATH, "tests.log"),
    ]

    @it.should("return hdl_checker version")
    @disableVunit
    def test():
        from hdl_checker import server
        from hdl_checker import __version__ as version

        with patch.object(server, "sys") as sys:
            with patch.object(
                server.argparse._sys, "argv", [p.abspath(server.__file__), "--version"]
            ):
                server.parseArguments()

        sys.stdout.write.assert_called_with(version + "\n")
        sys.exit.assert_called_with(0)

    with it.having("http server"):

        def startCodeCheckerServer():
            it._host = "127.0.0.1"
            it._port = str(_getUnusedLocalhostPort())
            it._url = "http://{0}:{1}".format(it._host, it._port)

            cmd = list(_SERVER_BASE_CMD) + ["--host", it._host, "--port", str(it._port)]

            _logger.info("Starting hdl_checker server with '%s'", " ".join(cmd))

            stdout_r, stdout_w = os.pipe()
            stderr_r, stderr_w = os.pipe()

            it.stdout = os.fdopen(stdout_r, "rb")
            it.stderr = os.fdopen(stderr_r, "rb")

            it._server = subp.Popen(
                cmd,
                env=os.environ.copy(),
                stdout=os.fdopen(stdout_w, "wb"),
                stderr=os.fdopen(stderr_w, "wb"),
            )
            waitForServer()

        def startServerAttachedToPid(pid):
            it._url = "http://{0}:{1}".format(it._host, it._port)

            cmd = list(_SERVER_BASE_CMD) + [
                "--host",
                it._host,
                "--port",
                str(it._port),
                "--attach-to-pid",
                str(pid),
            ]

            _logger.info("Starting hdl_checker server with '%s'", " ".join(cmd))

            it._server = subp.Popen(cmd, env=os.environ.copy())
            waitForServer()

        def waitForServer():
            event = Event()

            def wait():
                # Wait until the server is up and replying
                start = time.time()
                while not event.is_set():
                    try:
                        reply = requests.post(it._url + "/get_diagnose_info")
                        if reply.ok:
                            _logger.info(
                                "Server replied OK after %.1fs", time.time() - start
                            )
                            event.set()
                            return
                    except requests.ConnectionError:
                        pass
                    except:
                        _logger.exception(
                            "Exception while waiting for server to respond"
                        )
                        raise

                    time.sleep(0.5)

                _logger.info("Exiting wait thread")

            Thread(target=wait).start()
            # Wait 10s for the server to start responding
            event.wait(timeout=10)

            if event.is_set():
                return

            # Set the event from here to force the wait thread to exit
            event.set()

            _logger.error("Server is not replying")

            it._server.terminate()
            terminateProcess(it._server.pid)

            _logger.error("stderr: %s", it.stderr.read())

            it.fail("Server is not responding")

        def waitUntilBuildFinishes(data):
            _logger.info("Waiting for 30s until build is finished")
            for i in range(30):
                #  time.sleep(1)
                _logger.info("Elapsed %ds", i)
                _ = requests.post(it._url + "/get_messages_by_path", data)
                ui_messages = requests.post(it._url + "/get_ui_messages", data)
                _logger.debug("==> %s", ui_messages.json)
                if ui_messages.json["ui_messages"] == []:
                    _logger.info("Ok, done")
                    return

            assert False, "Server is still building after 30s"

        @it.has_teardown
        def teardown():
            it._server.terminate()
            terminateProcess(it._server.pid)

        @it.should("start and respond a request")  # type: ignore
        @disableVunit
        def test():
            startCodeCheckerServer()
            some_project = _path("some_project")
            open(some_project, "w").write("")
            # Ensure the server is active
            reply = requests.post(
                it._url + "/get_diagnose_info", data={"project_file": some_project}
            )
            it.assertTrue(reply.ok, "Reply was not OK: {}".format(reply))

        @it.should("shutdown the server when requested")  # type: ignore
        @disableVunit
        def test():
            # Send a request to the shutdown addr
            with it.assertRaises(requests.ConnectionError):
                reply = requests.post(it._url + "/shutdown")
                it.assertFalse(reply.ok)

            it._server.terminate()
            terminateProcess(it._server.pid)

        @it.should(  # type: ignore
            "terminate when the parent PID is not running anymore"
        )
        def test():

            queue = Queue()

            proc = Process(target=doNothing, args=(queue,))
            proc.start()

            _logger.info("Started dummy process with PID %d", proc.pid)
            startServerAttachedToPid(proc.pid)
            time.sleep(3)
            _logger.info("Allowing the dummy process to finish")
            queue.put(1)
            proc.join()

            if isProcessRunning(proc.pid):
                _logger.warning("Dummy process %d was still running", proc.pid)
                proc.terminate()
                time.sleep(1)
                it.assertFalse(
                    isProcessRunning(proc.pid),
                    "Process %d is still running after terminating " "it!" % proc.pid,
                )

            time.sleep(1)
            _logger.info("Server should have died by now")

            with it.assertRaises(requests.ConnectionError):
                requests.post(it._url + "/get_diagnose_info")

    with it.having("LSP server"):

        @it.should("initialize with no project file")  # type: ignore
        @disableVunit
        def test():
            client_server = _ClientServer()
            response = client_server.client.lsp.send_request(
                features.INITIALIZE,
                InitializeParams(
                    process_id=1234,
                    capabilities=ClientCapabilities(),
                    root_uri=uris.from_fs_path(TEST_TEMP_PATH),
                ),
            ).result(timeout=CALL_TIMEOUT)

            _logger.debug("Response: %s", response)
            it.assertEqual(response.capabilities.textDocumentSync, 2)
            it.assertEqual(response.capabilities.hoverProvider, True)

            shutdown_response = client_server.client.lsp.send_request(
                features.SHUTDOWN
            ).result(2)
            client_server.client.lsp.notify(features.EXIT)
            it.assertIsNone(shutdown_response)

        @it.should("log to temporary files if files aren't specified")  # type: ignore
        @disableVunit
        def test():
            from hdl_checker import server

            with patch.object(
                server.argparse._sys, "argv", [p.abspath(server.__file__)]
            ):
                args = server.parseArguments()

            it.assertIs(args.log_stream, server.sys.stdout)

            if ON_LINUX:
                it.assertEqual(
                    p.basename(args.stderr),
                    "hdl_checker_stderr_pid{}.log".format(os.getpid()),
                )
            else:
                it.assertTrue(
                    p.basename(args.stderr).startswith("hdl_checker_stderr_pid"),
                    "stderr log should not be {}".format(args.stderr),
                )

    with it.having("LSP server executable"):

        def assertCommandPrints(cmd, stdout, **kwargs):
            _logger.debug("Running command: %s", cmd)
            output = subp.check_output(cmd, **kwargs).decode().strip()
            it.assertEqual(output, stdout)

        @it.should("report version correctly")
        def test():
            assertCommandPrints(["hdl_checker", "--version"], hdl_checker.__version__)

        def startServerWrapper(cmd):
            log_file = tempfile.mktemp()

            actual_cmd = cmd + ["--log-stream", log_file]

            _logger.info("Actual command: %s", actual_cmd)

            proc = subp.Popen(
                actual_cmd, stdin=subp.PIPE, stdout=subp.PIPE, stderr=subp.PIPE
            )

            stdout, stderr = proc.communicate(LSP_SHUTDOWN, timeout=2)

            it.assertEqual(
                stdout, b"", "stdout should be empty but got\n{}".format(stdout)
            )

            it.assertEqual(
                stderr, b"", "stderr should be empty but got\n{}".format(stderr)
            )

            # On Windows the Popen PID and the *actual* PID don't always match
            # for some reason. Since we're not testing this, just skip the
            # first line
            log_content = open(log_file, "rb").read().decode().split("\n")

            expected = [
                "Starting server. Our PID is {}, no parent PID to attach to. "
                "Version string for hdl_checker is '{}'".format(
                    proc.pid, hdl_checker.__version__
                ),
                #  "Starting HdlCheckerLanguageServer IO language server",
            ]

            _logger.info(
                "Log content:\n-----\n%s\n-----\n",
                "\n".join(("> %s" % line for line in log_content)),
            )

            if ON_WINDOWS:
                log_content = log_content[1:]
                expected = expected[1:]

            for line in expected:
                it.assertIn(line, "\n".join(log_content))

            os.remove(log_file)

        @it.should(  # type: ignore
            "start server given the --lsp flag and setting stderr"
        )
        def test():
            startServerWrapper(
                [
                    "hdl_checker",
                    "--lsp",
                    "--stderr",
                    p.join(TEST_LOG_PATH, "hdl_checker_stderr.log"),
                ]
            )

        @it.should("start server given the --lsp flag")  # type: ignore
        def test():
            startServerWrapper(["hdl_checker", "--lsp"])

        @it.should("log to temporary files if files aren't specified")  # type: ignore
        @disableVunit
        def test():
            from hdl_checker import server

            with patch.object(
                server.argparse._sys, "argv", [p.abspath(server.__file__), "--lsp"]
            ):
                args = server.parseArguments()

            if ON_LINUX:
                it.assertEqual(
                    p.basename(args.log_stream),
                    "hdl_checker_log_pid{}.log".format(os.getpid()),
                )

                it.assertEqual(
                    p.basename(args.stderr),
                    "hdl_checker_stderr_pid{}.log".format(os.getpid()),
                )
            else:
                it.assertTrue(
                    p.basename(args.log_stream).startswith("hdl_checker_log_pid"),
                    "log file should not be {}".format(args.stderr),
                )
                it.assertTrue(
                    p.basename(args.stderr).startswith("hdl_checker_stderr_pid"),
                    "stderr file should not be {}".format(args.stderr),
                )

        @it.should("disable writing to log when passing --log-stream NONE")  # type: ignore
        @disableVunit
        def test():
            from hdl_checker import server

            # Check it works with LSP
            with patch.object(
                server.argparse._sys,
                "argv",
                [p.abspath(server.__file__), "--lsp", "--log-stream", "NONE"],
            ):
                args = server.parseArguments()

            it.assertIsNone(args.log_stream)

            # Check it works with HTTP server
            with patch.object(
                server.argparse._sys,
                "argv",
                [p.abspath(server.__file__), "--log-stream", "NONE"],
            ):
                args = server.parseArguments()

            it.assertIsNone(args.log_stream)

        @it.should("disable writing to stderr when passing --stderr NONE")  # type: ignore
        @disableVunit
        def test():
            from hdl_checker import server

            # Check it works with LSP
            with patch.object(
                server.argparse._sys,
                "argv",
                [p.abspath(server.__file__), "--lsp", "--stderr", "NONE"],
            ):
                args = server.parseArguments()

            it.assertIsNone(args.stderr)

            # Check it works with HTTP server
            with patch.object(
                server.argparse._sys,
                "argv",
                [p.abspath(server.__file__), "--stderr", "NONE"],
            ):
                args = server.parseArguments()

            it.assertIsNone(args.stderr)


@patch("hdl_checker.lsp.HdlCheckerLanguageServer.start_io")
@patch("hdl_checker.server._binaryStdio", return_value=("stdin", "stdout"))
@patch("hdl_checker.server._setupPipeRedirection")
def test_StartLsp(redirection, binary_stdio, start_server):
    args = type(
        "args",
        (object,),
        {"lsp": True, "stderr": "stderr", "log_stream": None, "attach_to_pid": None},
    )

    server.run(args)

    redirection.assert_called_once_with(None, "stderr")
    binary_stdio.assert_called_once()
    start_server.assert_called_once_with(stdin="stdin", stdout="stdout")


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"""
Test code inside hdl_checker.lsp that doesn't depend on a server/client setup
"""

import logging
import time
from multiprocessing import Queue
from tempfile import mkdtemp
from typing import Any

import mock
import parameterized  # type: ignore
import unittest2  # type: ignore
from pygls.types import DiagnosticSeverity, MessageType, Position, Range

from hdl_checker import lsp
from hdl_checker.diagnostics import CheckerDiagnostic, DiagType
from hdl_checker.path import Path, TemporaryPath
from hdl_checker.utils import debounce

_logger = logging.getLogger(__name__)


class TestCheckerDiagToLspDict(unittest2.TestCase):
    """
    Test code inside hdl_checker.lsp that doesn't depend on a server/client
    setup
    """

    @parameterized.parameterized.expand(
        [
            (DiagType.INFO, DiagnosticSeverity.Information),
            (DiagType.STYLE_INFO, DiagnosticSeverity.Information),
            (DiagType.STYLE_WARNING, DiagnosticSeverity.Information),
            (DiagType.STYLE_ERROR, DiagnosticSeverity.Information),
            (DiagType.WARNING, DiagnosticSeverity.Warning),
            (DiagType.ERROR, DiagnosticSeverity.Error),
            (DiagType.NONE, DiagnosticSeverity.Error),
        ]
    )
    def test_convertingDiagnosticType(self, diag_type, severity):
        # type: (...) -> Any
        """
        Test conversion between hdl_checker.DiagType to pygls.types.DiagnosticSeverity
        """
        _logger.info("Running %s and %s", diag_type, severity)

        diag = lsp.checkerDiagToLspDict(
            CheckerDiagnostic(
                checker="hdl_checker test",
                text="some diag",
                filename=Path("filename"),
                line_number=0,
                column_number=0,
                error_code="error code",
                severity=diag_type,
            )
        )

        self.assertEqual(diag.code, "error code")
        self.assertEqual(diag.source, "hdl_checker test")
        self.assertEqual(diag.message, "some diag")
        self.assertEqual(diag.severity, severity)
        self.assertEqual(
            diag.range,
            Range(
                start=Position(line=0, character=0), end=Position(line=0, character=1),
            ),
        )

    def test_workspaceNotify(self) -> None:  # pylint: disable=no-self-use
        """
        Test server notification messages call the appropriate LS methods
        """
        workspace = mock.Mock()
        workspace.show_message = mock.Mock()

        server = lsp.Server(
            workspace, root_dir=TemporaryPath(mkdtemp(prefix="hdl_checker_"))
        )
        workspace.show_message.reset_mock()

        server._handleUiInfo("some info")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with("some info", MessageType.Info)
        workspace.show_message.reset_mock()

        server._handleUiWarning("some warning")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with(
            "some warning", MessageType.Warning
        )
        workspace.show_message.reset_mock()

        server._handleUiError("some error")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with("some error", MessageType.Error)

    def test_debounceWithoutKey(self):
        _logger.info("#" * 100)
        interval = 0.1

        queue = Queue()

        def func(arg):
            _logger.info("Called with %s", arg)
            queue.put(arg)

        wrapped = debounce(interval)(func)
        self.assertTrue(queue.empty())

        wrapped(1)
        wrapped(2)
        self.assertTrue(queue.empty())

        time.sleep(2 * interval)

        self.assertEqual(queue.get(1), 2)
        self.assertTrue(queue.empty())

    def test_debounceWithKey(self):
        _logger.info("#" * 100)
        interval = 0.1

        obj = mock.Mock()

        def func(arg):
            _logger.info("Called with %s", arg)
            obj(arg)

        wrapped = debounce(interval, "arg")(func)
        obj.assert_not_called()

        wrapped(1)
        wrapped(2)
        wrapped(3)

        obj.assert_not_called()

        time.sleep(2 * interval)

        obj.assert_has_calls(
            [mock.call(1), mock.call(2), mock.call(3),], any_order=True
        )

        wrapped(4)
        wrapped(4)
        wrapped(4)
        time.sleep(2 * interval)

        obj.assert_has_calls(
            [mock.call(1), mock.call(2), mock.call(3), mock.call(4),], any_order=True
        )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring
# pylint: disable=useless-object-inheritance

import functools
import logging
import os
import os.path as p
import shutil
import subprocess as subp
import time
from multiprocessing import Queue
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import mock
import six
import unittest2  # type: ignore
from parameterized import parameterized_class  # type: ignore
from pygls import uris
from unittest2 import TestCase

from hdl_checker import exceptions
from hdl_checker.core import HdlCheckerCore
from hdl_checker.builders.base_builder import BaseBuilder
from hdl_checker.diagnostics import CheckerDiagnostic
from hdl_checker.parsers.elements.dependency_spec import RequiredDesignUnit
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import FileType
from hdl_checker.utils import ON_LINUX, ON_WINDOWS, removeDuplicates, samefile

_logger = logging.getLogger(__name__)


MockDep = Union[Tuple[str], Tuple[str, str]]


class DummyServer(HdlCheckerCore):
    "Class for testing HdlCheckerCore"
    _server_index = 0

    def __init__(self, *args, **kwargs):
        _logger.info("Creating server %d", DummyServer._server_index)
        self._msg_queue = Queue()  # type: Queue[Tuple[str, str]]
        self._ui_handler = logging.getLogger("server %d/UI" % DummyServer._server_index)
        DummyServer._server_index += 1
        super(DummyServer, self).__init__(*args, **kwargs)

    def _handleUiInfo(self, message):
        self._msg_queue.put(("info", message))
        self._ui_handler.info("[UI INFO]: %s", message)

    def _handleUiWarning(self, message):
        self._msg_queue.put(("warning", message))
        self._ui_handler.info("[UI WARNING]: %s", message)

    def _handleUiError(self, message):
        self._msg_queue.put(("error", message))
        self._ui_handler.info("[UI ERROR]: %s", message)

    def getUiMessages(self):
        while not self._msg_queue.empty():
            yield self._msg_queue.get()


class SourceMock(object):
    _logger = logging.getLogger("SourceMock")
    base_path = ""

    def __init__(
        self,
        design_units,  # type: Iterable[Dict[str, str]]
        library=None,  # type: str
        dependencies=None,  # type: Iterable[MockDep]
        filename=None,  # type: Optional[str]
    ):

        self._design_units = list(design_units or [])

        if filename is not None:
            self._filename = Path(p.join(self.base_path, filename))
        else:
            library = "lib_not_set" if library is None else library
            self._filename = Path(
                p.join(
                    self.base_path,
                    library,
                    "_{}.vhd".format(self._design_units[0]["name"]),
                )
            )

        self.filetype = FileType.fromPath(self.filename)
        #  self.abspath = p.abspath(self.filename)
        self.flags = []  # type: ignore

        self.library = library
        self._dependencies = []  # type: List[RequiredDesignUnit]
        for dep_spec in dependencies or []:
            _name = dep_spec[0]
            _library = "work"

            try:
                _library, _name = dep_spec  # type: ignore
            except ValueError:
                pass

            self._dependencies.append(
                RequiredDesignUnit(
                    self._filename,
                    Identifier(_name, False),
                    Identifier(_library, False),
                )
            )

        self._createMockFile()

    @property
    def filename(self):
        # type: () -> Path
        return self._filename

    def getLibraries(self):
        return removeDuplicates([x.library for x in self._dependencies])

    def _createMockFile(self):
        # type: () -> None
        self._logger.debug("Creating mock file: %s", self)
        libs = self.getLibraries()

        lines = []

        for lib in libs:
            lines.append("library {0};".format(lib.display_name))

        for dependency in self._dependencies:
            if dependency.library is not None:
                dependency_name = dependency.library.display_name
            else:
                dependency_name = "work"
            lines.append(
                "use {0}.{1};".format(dependency_name, dependency.name.display_name)
            )

        # Separate if there was libraries already added
        if lines:
            lines.append("")

        for design_unit in self._design_units:
            type_ = design_unit["type"]
            name = design_unit["name"]

            lines.append("{0} {1} is".format(type_, name))
            lines.append("")
            lines.append("end {0} {1};".format(type_, name))
            lines.append("")

        for i, line in enumerate(lines, 1):
            self._logger.debug("%2d | %s", i, line)

        try:
            os.makedirs(p.dirname(self.filename.name))
        except OSError:
            pass

        with open(self.filename.name, "w") as fd:
            fd.write("\n".join(lines))

    def __repr__(self):
        return (
            "{}(library='{}', design_units={}, dependencies={}, "
            "filename={}".format(
                self.__class__.__name__,
                self.library,
                self._design_units,
                self._dependencies,
                self.filename,
            )
        )

    def getmtime(self):
        return p.getmtime(self.filename)

    def getDesignUnits(self):
        return self._design_units

    def getDependencies(self):
        return self._dependencies

    def getRawSourceContent(self):
        return open(self.filename).read()


class MockBuilder(BaseBuilder):  # pylint: disable=abstract-method
    _logger = logging.getLogger("MockBuilder")
    builder_name = "mock_builder"
    file_types = (FileType.vhdl,)

    def __init__(self, work_folder, *args, **kwargs):
        # type: (...) -> None
        self._work_folder = work_folder  # type: Path
        if not p.exists(self._work_folder.name):
            os.makedirs(self._work_folder.name)

        super(MockBuilder, self).__init__(work_folder, *args, **kwargs)

    def _makeRecords(self, _):  # pragma: no cover
        return []

    def _shouldIgnoreLine(self, line):  # pragma: no cover
        return True

    def _checkEnvironment(self):
        return

    @staticmethod
    def isAvailable():
        return True

    def _buildSource(self, path, library, flags=None):
        self._logger.debug(
            "Building path=%s, library=%s, flags=%s", path, library, flags
        )
        return [], []

    def _createLibrary(self, library):  # pylint: disable=unused-argument
        pass

    def _parseBuiltinLibraries(self):
        # type: (...) -> Any
        return (
            Identifier("ieee", case_sensitive=False),
            Identifier("std", case_sensitive=False),
        )


class FailingBuilder(MockBuilder):  # pylint: disable=abstract-method
    _logger = logging.getLogger("FailingBuilder")
    builder_name = "FailingBuilder"

    def _checkEnvironment(self):
        raise exceptions.SanityCheckError(self.builder_name, "Fake error")


disableVunit = mock.patch("hdl_checker.builder_utils.foundVunit", lambda: False)


class PatchBuilder(object):
    def __init__(self, meth=None):
        def getBuilderByName(name):
            "Returns the builder class given a string name"
            from hdl_checker.builders.msim import MSim
            from hdl_checker.builders.ghdl import GHDL
            from hdl_checker.builders.xvhdl import XVHDL
            from hdl_checker.builders.fallback import Fallback

            # Check if the builder selected is implemented and create the
            # builder attribute
            _logger.info("Getting builder class for %s", name)
            if name == MockBuilder.builder_name:
                return MockBuilder
            if name == "msim":
                return MSim
            if name == "xvhdl":
                return XVHDL
            if name == "ghdl":
                return GHDL

            return Fallback

        self.meth = meth
        self.patches = (
            mock.patch(
                "hdl_checker.core.getPreferredBuilder",
                side_effect=[MockBuilder,],
            ),
            mock.patch("hdl_checker.core.getBuilderByName", getBuilderByName),
            disableVunit,
        )

    def __enter__(self):
        _logger.info("Starting patches")
        list(x.start() for x in self.patches)

    def __exit__(self, *args, **kwargs):
        _logger.info("Stopping patches")
        list(x.stop() for x in self.patches)

    def __call__(self, *args, **kwargs):
        self.__enter__()
        try:
            if self.meth is not None:
                yield self.meth()
            else:
                yield None
        except:
            _logger.exception("Failed to run %s", self.meth)
            raise
        finally:
            self.__exit__()


def sanitizePath(*args):
    """
    Join args with pathsep, gets the absolute path and normalizes
    """
    return p.normpath(
        p.abspath(p.join(*args))  # pylint: disable=no-value-for-parameter
    )


def assertSameFile(it):  # pylint: disable=invalid-name
    def wrapper(first, second, msg=None):
        try:
            os.stat(first)
        except TypeError:
            it.fail("Invalid first argument of type {}".format(type(first)))
        try:
            os.stat(second)
        except TypeError:
            it.fail("Invalid first argument of type {}".format(type(second)))
        if not samefile(p.abspath(first), p.abspath(second)):
            _msg = "" if msg is None else "{}\n".format(msg)
            it.fail(
                "{}Paths '{}' and '{}' differ".format(
                    _msg, p.abspath(first), p.abspath(second)
                )
            )

    return wrapper


#  def assertCountEqual(it):  # pylint: disable=invalid-name

#      assert six.PY2, "Only needed on Python2"

#      def wrapper(first, second, msg=None):
#          temp = list(second)  # make a mutable copy
#          not_found = []
#          for elem in first:
#              try:
#                  temp.remove(elem)
#              except ValueError:
#                  not_found.append(elem)

#          error_details = []

#          if not_found:
#              error_details += [
#                  "Second list is missing item {}".format(x) for x in not_found
#              ]

#          error_details += ["First list is missing item {}".format(x) for x in temp]

#          if error_details:
#              # Add user message at the top
#              error_details = [msg] + error_details
#              error_details += ["", "Lists {} and {} differ".format(first, second)]
#              it.fail("\n".join([str(x) for x in error_details]))

#      return wrapper


def writeListToFile(filename, _list):  # pragma: no cover
    "Well... writes '_list' to 'filename'. This is for testing only"
    # Wait a little bit to force the timestamp rad via os.path.getmtime to
    # change
    time.sleep(0.1)
    open(filename, mode="w").write("\n".join([str(x) for x in _list]))

    for i, line in enumerate(_list):
        _logger.debug("%2d | %s", i + 1, line)


if not ON_WINDOWS:
    TEST_ENVS = {
        "ghdl": os.environ["GHDL_PATH"],
        "msim": os.environ["MODELSIM_PATH"],
        "xvhdl": os.environ["XSIM_PATH"],
        "fallback": None,
    }
else:
    TEST_ENVS = {"fallback": None}


def parametrizeClassWithBuilders(cls):
    cls.assertSameFile = assertSameFile(cls)

    keys = ["builder_name", "builder_path"]
    values = []
    for name, path in TEST_ENVS.items():
        values += [(name, path)]

    return parameterized_class(keys, values)(cls)


def getTestTempPath(name):
    # type: (str) -> str
    name = name.replace(".", "_")
    path = p.abspath(p.join(os.environ["TOX_ENV_DIR"], "tmp", name))

    # Create a path for each test to allow running multiple tests concurrently
    i = 0
    while p.exists(path):
        path = p.abspath(p.join(os.environ["TOX_ENV_DIR"], "tmp", "%s_%d" % (name, i)))
        i += 1

    if not p.exists(path):
        os.makedirs(path)
    return path


def setupTestSuport(path):
    # type: (str) -> None
    """Copy contents of .ci/test_support_path/ to the given path"""
    _logger.info("Setting up test support at %s", path)

    test_support_path = os.environ["CI_TEST_SUPPORT_PATH"]

    paths_to_copy = os.listdir(test_support_path)

    # Create the parent directory
    if not p.exists(p.dirname(path)):
        _logger.info("Creating %s", p.dirname(path))
        os.makedirs(p.dirname(path))

    for path_to_copy in paths_to_copy:
        src = p.join(test_support_path, path_to_copy)
        dest = p.join(path, path_to_copy)

        if p.exists(dest):
            _logger.info("Destination path %s already exists, removing it", dest)
            shutil.rmtree(dest)

        _logger.info("Copying %s to %s", src, dest)
        shutil.copytree(src, dest)


def logIterable(msg, iterable, func):
    # type: (str, Iterable[Any], Callable) -> None
    func(msg)
    for i, item in enumerate(iterable, 1):
        func("- {:2d} {}".format(i, item))


def windowsOnly(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not ON_WINDOWS:
            return unittest2.skip("Windows only test")
        return func(*args, **kwargs)

    return wrapper


def linuxOnly(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not ON_LINUX:
            return unittest2.skip("Linux only test")
        return func(*args, **kwargs)

    return wrapper


def toCheckerDiagnostic(uri: str, diags: Any) -> Iterable[CheckerDiagnostic]:
    for diag in diags:
        yield CheckerDiagnostic(
            text=diag.message,
            checker=diag.source,
            filename=uris.to_fs_path(uri),
            line_number=diag.range.start.line,
            column_number=diag.range.start.character,
            error_code=diag.code,
            severity="Error",
        )
